import	re
import	datetime
import  base64
import  hashlib
import  hmac
import	urllib
import	urlparse

from	django.shortcuts		import render_to_response

from	sharedapps.turbo_oauth		import models


_TEMPORARY_TOKEN_LIFETIME	= 3600		# Maximum lifespan of a temporary token, in seconds
_TIMESTAMP_WINDOW_RADIUS	= 1800		# Maxmimum allowable delta, in seconds, between client and server clocks
_MIN_NONCE_LENGTH		= 16		# Minimum legal nonce length
_MAX_NONCE_LENGTH		= 64		# Maxmimum legal nonce length


class Http400(Exception):
    def get_response(self):
        r = render_to_response('400.html', {'e':self})
        r.status_code = 400
        return r

class Http401(Exception):
    def get_response(self):
        r = render_to_response('401.html', {'e':self})
        r.status_code = 401
        return r

class _Http401_soft(Http400):
    # For our current application, we don't want to return a lot of 401s
    # This exception class lets us replace (for now) 401s with 400s
    # One quick change to its parent will let us remove this hack
    pass

class OAuthData():
    def __init__(self, all_fields, signable_fields, consumer, credential):
        self.all_fields		= all_fields
        self.signable_fields	= signable_fields
        self.consumer		= consumer
        self.credential		= credential
        self.user		= None


def oauth_user_required(f):
    def i(request, *args, **kwargs):
        try:
            _process_oauth_request(request)
            if not request.oauth_data:
                raise _Http401_soft('Bad authorization header')
            if not request.oauth_data.user:
                raise Http401('No user')
            return f(request, *args, **kwargs)
        except (Http400, Http401), e:
            return e.get_response()
    return i

def oauth_only_consumer_required(f):
    def i(request, *args, **kwargs):
        try:
            _process_oauth_request(request)
            if not request.oauth_data:
                raise _Http401_soft('Bad authorization header')
            return f(request, *args, **kwargs)
        except (Http400, Http401), e:
            return e.get_response()
    return i


# Encode a string as per OAuth rules
def _encode(s):
    return urllib.quote(s.encode('utf-8'), '~')

# Turn a list of key-list pairs into a list of key-value pairs
def _explode_multipairs(p):
    rv = []
    for k,l in p:
        for v in l:
            rv.append((k,v))
    return rv


# Perform common tests and data extraction
def _process_authorization_header(s):
    legal_parameters =		set(('realm',
                                     'oauth_consumer_key',
                                     'oauth_token',
                                     'oauth_signature_method',
                                     'oauth_timestamp',
                                     'oauth_nonce',
                                     'oauth_version',
                                     'oauth_verifier',
                                     'oauth_callback',
                                     'oauth_signature'))
    optional_parameters =	set(('realm',
                                     'oauth_token',
                                     'oauth_version',
                                     'oauth_verifier',
                                     'oauth_callback'))
    unsigned_parameters =	set(('realm',
                                     'oauth_signature'))

    # Special check for scheme (case-insensitive)
    if s[:6].lower() != 'oauth ': return None

    # Parse the header
    rv = {}
    for k,v in re.findall('\s*(.+?)="(.*?)"(?:,|$)', s[6:]):
        # Decode
        k = urllib.unquote(k).decode('utf-8')
        v = urllib.unquote(v).decode('utf-8')

        # Special check for realm (case-insensitive)
        if k.lower() == 'realm': k = k.lower()

        # Check for unsupported parameters
        if k not in legal_parameters: raise Http400('Unexpected parameter: '+k)

        # Check for duplicate parameters
        if k in rv: raise Http400('Duplicate parameter: '+k)

        # Store
        rv[k] = v

    # Check for missing parameters
    if legal_parameters - optional_parameters - set(rv.keys()): raise Http400('Missing parameter')

    # Check for bad version (if specified)
    if ('oauth_version' in rv) and (rv['oauth_version'] != '1.0'): raise Http400('Bad version')

    # Check for bad signature method
    if rv['oauth_signature_method'] != 'HMAC-SHA1': raise Http400('Bad signature method')

    # Check for bad timestamp
    timestamp = datetime.datetime.fromtimestamp(int(rv['oauth_timestamp']))
    timestamp_delta = datetime.datetime.now() - timestamp
    if abs(timestamp_delta.days*24*3600 + timestamp_delta.seconds) > _TIMESTAMP_WINDOW_RADIUS: raise _Http401_soft('Bad timestamp')

    # Check for bad nonce
    nonce_length = len(rv['oauth_nonce'])
    if (nonce_length < _MIN_NONCE_LENGTH) or (nonce_length > _MAX_NONCE_LENGTH): raise _Http401_soft('Bad nonce')

    # Find consumer (or error)
    try:
        consumer = models.Consumer.objects.get(key=rv['oauth_consumer_key'])
    except models.Consumer.DoesNotExist:
        raise _Http401_soft('Unknown consumer')

    # Find credential (or error) iff specified
    if 'oauth_token' in rv:
        try:
            credential = models.Credential.objects.get(token=rv['oauth_token'])
        except models.Credential.DoesNotExist:
            raise Http401('Unknown credential')
    else:
        credential = None

    # Check for credential/consumer mismatch
    if credential and (credential.consumer != consumer): raise Http401('Client/token credential mismatch')

    # Check for replay
    if models.ReplayBar.objects.filter(token=credential).filter(timestamp=timestamp).filter(nonce=rv['oauth_nonce']).count():
        raise _Http401_soft('Replay')

    # Delete old replay records
    cutoff = datetime.datetime.now() + datetime.timedelta(seconds=-_TIMESTAMP_WINDOW_RADIUS)
    models.ReplayBar.objects.filter(token=credential).filter(timestamp__lt=cutoff).delete()
    
    # Add a new replay record
    models.ReplayBar(token=credential, timestamp=timestamp, nonce=rv['oauth_nonce']).save()

    # Return a data structure
    return OAuthData(rv, [(k,v) for k,v in rv.items() if k not in unsigned_parameters], consumer, credential)


# Perform common OAuth request handling
def _process_oauth_request(request):
    # Get Authorization parameters (stop processing if none were retrieved)
    request.oauth_data = _process_authorization_header(request.META.get('HTTP_AUTHORIZATION', ''))
    if not request.oauth_data: return

    # Reject expired temporary tokens
    if ((request.oauth_data.credential != None)		and
        (not request.oauth_data.credential.is_token())	and
        (request.oauth_data.credential.issued_on+datetime.timedelta(seconds=_TEMPORARY_TOKEN_LIFETIME) < datetime.datetime.now())):
        raise Http401('Expired temporary token')

    # Get an UC method
    method = request.method

    # Get a scheme (and the default port)
    if request.is_secure():
        scheme = 'https://'
        defport = 443
    else:
        scheme = 'http://'
        defport = 80

    # Get the authority as declared by the HTTP "Host" request header field (and the actual port)
    if request.META['HTTP_HOST'].find(':') != -1:
        host, port = request.META['HTTP_HOST'].split(':'); port = int(port)
    else:
        host, port = request.META['HTTP_HOST'], defport

    # Get the path of the URI, appropriately quoted (the '~' might be controversial)
    # Note that the encoding here just recreates what went over the wire; a subsequent 2nd encoding is necessary for SBS construction
    path = urllib.quote(request.path.encode('utf-8'), '~/')

    # Get query string parameters
    qs = urlparse.parse_qsl(request.META.get('QUERY_STRING', ''), keep_blank_values=True)
    qs = [(k.decode('utf-8'), v.decode('utf-8')) for k,v in qs]

    # Get entity-body parameters
    # Not quite RFC5849 compliant, because GET entity-bodies are *always* ignored, but:
    # * They're very uncommon
    # * Arguably they have no semantic meaning, and need not be signed
    # * Django (or Apache) drops them, or at least doesn't make them readily available
    # * Let's not talk about DELETEs, PUTs, etc.
    body = []
    if (method == 'POST') and (request.META.get('CONTENT_TYPE', '') == 'application/x-www-form-urlencoded'):
        body = _explode_multipairs(request.POST.lists())

    # Build SBS
    if port == defport:
        uri = scheme + host.lower() + path
    else:
        uri = scheme + host.lower() + (':%d'%port) + path
    params = '&'.join('%s=%s'%(k,v) for k,v in sorted((_encode(k),_encode(v)) for k,v in request.oauth_data.signable_fields+qs+body))
    sbs =  '&'.join(map(_encode, [method, uri, params]))

    # Build key
    key = _encode(request.oauth_data.consumer.secret) + '&'
    if request.oauth_data.credential:
        key = key + _encode(request.oauth_data.credential.secret)

    # Check signature
    if base64.b64encode(hmac.new(key, sbs, hashlib.sha1).digest()) != request.oauth_data.all_fields['oauth_signature']:
        raise _Http401_soft('Bad signature')

    # Assign a user, if possible
    if request.oauth_data.credential and request.oauth_data.credential.is_token():
        request.oauth_data.user = request.oauth_data.credential.user
